# -*- coding: utf-8 -*-
from django.conf import settings

BASECRM_API_URL = getattr(settings, 'BASECRM_API_URL', 'https://api.getbase.com/v2/')
BASECRM_API_KEY = getattr(settings, 'BASECRM_API_KEY', None)
BASECRM_USER_AGENT = getattr(settings, 'BASECRM_USER_AGENT', 'YunoJuno/1.0')
BASECRM_CACHE_USERS = getattr(settings, 'BASECRM_CACHE_USERS', True)
BASECRM_CACHE_STAGES = getattr(settings, 'BASECRM_CACHE_STAGES', True)
BASECRM_CACHE_PIPELINE = getattr(settings, 'BASECRM_CACHE_PIPELINE', True)
BASECRM_CACHE_AT_STARTUP = getattr(settings, 'BASECRM_CACHE_AT_STARTUP', True)


# -*- coding: utf-8 -*-
from django.apps import AppConfig

from . import (
    utils as base_api,
    exceptions as base_exceptions,
    settings as base_settings
)
from helpers import (
    get_pipelines_from_api,
    get_stages_from_api,
    get_users_from_api
)

class BaseCRMConfig(AppConfig):
    name = 'django_basecrm'
    verbose_name = "BaseCRM API client SDK for Django"
    pipeline = None
    stages = None
    users = None

    def ready(self):
        super(BaseCRMConfig, self).ready()
        if base_settings.BASECRM_CACHE_AT_STARTUP:
            self.instantiate_objects()

    def instantiate_objects(self, force=False):
        """
        Currently we assume a single pipeline (.get_pipelines() will error if there are > 1), which
        makes keeping track of stages easier. Deals progress through stages, and sending incorrect
        user IDs can lose deals, so it's useful to have a list available
        """
        self.instantiate_pipeline(force)
        self.instantiate_stages(force)
        self.instantiate_users(force)

    def instantiate_pipeline(self, force=False):
        if force is True or self.pipeline is None:
            p = get_pipelines_from_api()
            try:
                self.pipeline = p[0]
            except:
                self.pipeline = p

    def instantiate_stages(self, force=False):
        if force is True or self.stages is None:
            kwargs = {}
            if self.pipeline is not None:
                kwargs['pipeline_id'] = self.pipeline['id']
            self.stages = get_stages_from_api(**kwargs)

    def instantiate_users(self, force=False):
        if force is True or self.users is None:
            self.users = get_users_from_api()




# -*- coding: utf-8 -*-

from django.apps import apps as django_apps

from . import (
    utils as base_api,
    exceptions as base_exceptions,
    settings as base_settings,
)

"""
All functions at this level are simple wrappers that accept (as kwargs) any extra GET vars to be
sent to the API call. In addition, the special 'id' kwarg parameter will affect the API endpoint
URL itself instead of being appended as a GET var.
"""

def get_contacts(**kwargs):
    """
    Hits the API for contacts. If given an 'id' kwarg it will request that specific ID contact;
    any other kwargs will be passed to the API as GET params; e.g. page=2.

    Returns a list of contact dicts unless the 'id' param is set in which case it returns a dict
    """
    resp = base_api.request(base_api.RETRIEVE, 'contacts', kwargs)
    return base_api.parse(resp)

def create_contact(contact_dict):
    """
    Runs local validation on the given dict and gives passing ones to the API to create
    """
    if base_api.validate_contact_dict(base_api.CREATE, contact_dict):
        resp = base_api.request(base_api.CREATE, 'contacts', None, data=contact_dict)
        return base_api.parse(resp)
    else:
        # validation failed but the exception was suppressed
        pass

def update_contact(id, contact_dict):
    """
    Runs local validation on the given dict and gives passing ones to the API to update
    """
    if base_api.validate_contact_dict(base_api.UPDATE, contact_dict, skip_id=True):
        resp = base_api.request(base_api.UPDATE, 'contacts', {'id': id}, data=contact_dict)
        return base_api.parse(resp)
    else:
        # validation failed but the exception was suppressed
        pass

def get_deals(**kwargs):
    """
    Hits the API for deals. If given an 'id' kwarg it will request that specific ID deal;
    any other kwargs will be passed to the API as GET params; e.g. page=2.

    Returns a list of deal dicts unless the 'id' param is set in which case it returns a dict
    """
    resp = base_api.request(base_api.RETRIEVE, 'deals', kwargs)
    return base_api.parse(resp)

def create_deal(deal_dict):
    """
    Runs local validation on the given dict and gives passing ones to the API to create
    """
    if base_api.validate_deal_dict(base_api.CREATE, deal_dict):
        resp = base_api.request(base_api.CREATE, 'deals', None, data=deal_dict)
        return base_api.parse(resp)
    else:
        # validation failed but the exception was suppressed
        pass

def update_deal(id, deal_dict):
    """
    Runs local validation on the given dict and gives passing ones to the API to update
    """
    if base_api.validate_deal_dict(base_api.UPDATE, deal_dict, skip_id=True):
        resp = base_api.request(base_api.UPDATE, 'deals', {'id': id}, data=deal_dict)
        return base_api.parse(resp)
    else:
        # validation failed but the exception was suppressed
        pass

def get_notes(resource_type=None, resource_id=None, **kwargs):
    """
    Hits the API for notes. If resource_type and/or resource_id are
    specified they will be added to the query as GET params
    """
    if resource_type is not None:
        if resource_type not in ['lead', 'contact', 'deal']:
            raise base_exceptions.BaseCRMValidationError('Invalid resource type')
        else:
            kwargs['resource_type'] = resource_type

    if resource_id is not None:
        if resource_type is None:
            raise base_exceptions.BaseCRMValidationError('Resource type required when specifying resource ID')
        else:
            kwargs['resource_id'] = resource_id

    resp = base_api.request(base_api.RETRIEVE, 'notes', kwargs)
    return base_api.parse(resp)

def get_pipelines():
    """
    Note that we don't expect these to change often, so we are essentially caching this for the
    duration (there's no cachebusting)
    """
    if base_settings.BASECRM_CACHE_PIPELINE:
        app_conf = django_apps.get_app_config('django_basecrm')
        app_conf.instantiate_pipeline()
        pipeline = app_conf.pipeline
    else:
        pipeline = get_pipelines_from_api()
    return pipeline

def get_stages():
    """
    Note that we don't expect these to change often, so we are essentially caching this for the
    duration (there's no cachebusting)
    """
    if base_settings.BASECRM_CACHE_STAGES:
        app_conf = django_apps.get_app_config('django_basecrm')
        app_conf.instantiate_stages()
        stages = app_conf.stages
    else:
        stages = get_stages_from_api()
    return stages

def get_users():
    """
    Note that we don't expect these to change often, so we are essentially caching this for the
    duration (there's no cachebusting)
    """
    if base_settings.BASECRM_CACHE_USERS:
        app_conf = django_apps.get_app_config('django_basecrm')
        app_conf.instantiate_users()
        users = app_conf.users
    else:
        users = get_users_from_api()
    return users

def get_user_ids():
    return [x['id'] for x in get_users()]

def get_stage_ids():
    return [x['id'] for x in get_stages()]

def get_pipelines_from_api(**kwargs):
    """
    This is the API method, called by the appConfig.instantiate method
    """
    resp = base_api.request(base_api.RETRIEVE, 'pipelines', kwargs)
    if base_api.count(resp) > 1:
        raise NotImplementedError("We currently only cater for a single pipeline in BaseCRM")
    return base_api.parse(resp)

def get_stages_from_api(**kwargs):
    """
    This is the API method, called by the appConfig.instantiate method
    """
    resp = base_api.request(base_api.RETRIEVE, 'stages', kwargs)
    return base_api.parse(resp)

def get_users_from_api(**kwargs):
    """
    This is the API method, called by the appConfig.instantiate method
    """
    resp = base_api.request(base_api.RETRIEVE, 'users', kwargs)
    return base_api.parse(resp)



# -*- coding: utf-8 -*-
import mock
import types

from django.apps import apps as django_apps
from django.test import TestCase
from django.db.models.base import ModelBase

from . import apps, exceptions, helpers, serializers, utils


class RequestWrapperTests(TestCase):

    @mock.patch('%s.utils.base_settings' % __name__)
    def test_build_api_endpoint(self, settings):
        # bad api_urls aren't handled - caveat emptor
        settings.BASECRM_API_URL = 'api.base.com'
        params = None
        endpoint = 'contacts'
        result = utils._build_api_endpoint(endpoint, params)
        self.assertEqual(result, 'api.base.comcontacts')

        settings.BASECRM_API_URL = 'http://api.base.com/test'
        params = None
        endpoint = 'contacts'
        result = utils._build_api_endpoint(endpoint, params)
        self.assertEqual(result, 'http://api.base.com/testcontacts')

        settings.BASECRM_API_URL = 'http://api.base.com/test/'

        params = None
        endpoint = 'contacts'
        result = utils._build_api_endpoint(endpoint, params)
        self.assertEqual(result, 'http://api.base.com/test/contacts')

        params = {'id': 34}
        endpoint = 'contacts'
        result = utils._build_api_endpoint(endpoint, params)
        self.assertEqual(result, 'http://api.base.com/test/contacts/34')

        # only id params are used, and then only to add to the endpoint
        params = {'hello': 'world'}
        endpoint = 'contacts'
        result = utils._build_api_endpoint(endpoint, params)
        self.assertEqual(result, 'http://api.base.com/test/contacts')

        params = {'hello': 'world', 'id': 34}
        endpoint = 'contacts'
        result = utils._build_api_endpoint(endpoint, params)
        self.assertEqual(result, 'http://api.base.com/test/contacts/34')

        # but they're also not inspected - caveat emptor
        params = {'hello': 'world', 'id': 'world'}
        endpoint = 'contacts'
        result = utils._build_api_endpoint(endpoint, params)
        self.assertEqual(result, 'http://api.base.com/test/contacts/world')

    @mock.patch('%s.utils.base_settings' % __name__)
    def test_build_headers(self, settings):
        settings.BASECRM_API_KEY = 'an-api_key/that1s4lphaNuMer1c'
        settings.BASECRM_USER_AGENT = 'yourApp/1.0'

        extras = None
        result = utils._build_headers(extras)
        self.assertEqual(
            result,
            {
                'Accept': 'application/json',
                'Content-Type': 'application/json',
                'Authorization': 'Bearer an-api_key/that1s4lphaNuMer1c',
                'User-Agent': 'yourApp/1.0'
            }
        )

        extras = {
            'X-Custom-Header': 'test value'
        }
        result = utils._build_headers(extras)
        self.assertEqual(
            result,
            {
                'Accept': 'application/json',
                'Content-Type': 'application/json',
                'Authorization': 'Bearer an-api_key/that1s4lphaNuMer1c',
                'User-Agent': 'yourApp/1.0',
                'X-Custom-Header': 'test value'
            }
        )

        extras = {
            'Accept': 'text/html'
        }
        result = utils._build_headers(extras)
        self.assertEqual(
            result,
            {
                'Accept': 'text/html',
                'Content-Type': 'application/json',
                'Authorization': 'Bearer an-api_key/that1s4lphaNuMer1c',
                'User-Agent': 'yourApp/1.0'
            }
        )

        extras = {
            'Accept': 'text/html',
            'Content-Type': 'something',
            'Authorization': 'something else',
            'User-Agent': 'yourApp/2.0',
            'X-Custom-Header': 'test value'
        }
        result = utils._build_headers(extras)
        self.assertEqual(
            result,
            {
                'Accept': 'text/html',
                'Content-Type': 'something',
                'Authorization': 'something else',
                'User-Agent': 'yourApp/2.0',
                'X-Custom-Header': 'test value'
            }
        )

    @mock.patch('%s.utils.requests' % __name__)
    @mock.patch('%s.utils._build_api_endpoint' % __name__)
    @mock.patch('%s.utils._build_headers' % __name__)
    def test_request_wrapper(self, _b_headers, _b_endpoint, requests):
        response = mock.Mock()
        response.status_code = 200
        response.url = 'http://google.com/q=helloworld'
        requests.request.return_value = response
        _b_headers.return_value = {'Accept': 'application/json', 'User-Agent': 'Test'}
        _b_endpoint.return_value = 'http://google.com/'

        method = 'GET'
        endpoint = 'http://api.base.com/contacts'
        get_params = {}
        kwargs = {}
        result = utils._request(method, endpoint, get_params, **kwargs)
        self.assertEqual(result, response)
        _b_headers.assert_called_once_with(None)
        _b_endpoint.assert_called_once_with(endpoint, get_params)
        requests.request.assert_called_once_with(
            method,
            _b_endpoint.return_value,
            headers=_b_headers.return_value,
            params=get_params
        )

        _b_headers.reset_mock()
        _b_endpoint.reset_mock()
        requests.request.reset_mock()

        method = 'POST'
        endpoint = 'http://api.base.com/deals'
        get_params = {
            'test': True
        }
        kwargs = {
            'headers': {'Accept': 'plain/txt', 'User-Agent': 'BaseCRM Client'}
        }
        result = utils._request(method, endpoint, get_params, **kwargs)
        self.assertEqual(result, response)
        _b_headers.assert_called_once_with({'Accept': 'plain/txt', 'User-Agent': 'BaseCRM Client'})
        _b_endpoint.assert_called_once_with(endpoint, get_params)
        requests.request.assert_called_once_with(
            method,
            _b_endpoint.return_value,
            headers=_b_headers.return_value,
            params=get_params
        )

        _b_headers.reset_mock()
        _b_endpoint.reset_mock()
        requests.request.reset_mock()

        method = 'POST'
        endpoint = 'http://api.base.com/deals'
        get_params = {
            'test': True
        }
        kwargs = {
            'headers': {'Accept': 'plain/txt'},
            'some_other_requests_param': 87
        }
        result = utils._request(method, endpoint, get_params, **kwargs)
        self.assertEqual(result, response)
        _b_headers.assert_called_once_with({'Accept': 'plain/txt'})
        _b_endpoint.assert_called_once_with(endpoint, get_params)
        requests.request.assert_called_once_with(
            method,
            _b_endpoint.return_value,
            headers=_b_headers.return_value,
            params=get_params,
            some_other_requests_param=87
        )


class ValidationTests(TestCase):

    def test_validate_contact_dict(self):
        operation = utils.CREATE
        contact_dict = {}
        with self.assertRaises(exceptions.BaseCRMValidationError):
            utils.validate_contact_dict(operation, contact_dict)

        operation = utils.UPDATE
        contact_dict = {}
        with self.assertRaises(exceptions.BaseCRMValidationError):
            utils.validate_contact_dict(operation, contact_dict)

        operation = utils.CREATE
        contact_dict = {
            'is_organization': True,
            'name': 'Acme Carburretors Inc.'
        }
        result = utils.validate_contact_dict(operation, contact_dict)
        self.assertTrue(result)

        operation = utils.UPDATE
        contact_dict = {
            'is_organization': True,
            'name': 'Acme Carburretors Inc.'
        }
        with self.assertRaises(exceptions.BaseCRMValidationError):
            utils.validate_contact_dict(operation, contact_dict)

        operation = utils.CREATE
        contact_dict = {
            'is_organization': True,
            'first_name': 'Albert',
            'last_name': 'Einstein'
        }
        with self.assertRaises(exceptions.BaseCRMValidationError):
            utils.validate_contact_dict(operation, contact_dict)

        operation = utils.UPDATE
        contact_dict = {
            'is_organization': True,
            'first_name': 'Albert',
            'last_name': 'Einstein'
        }
        with self.assertRaises(exceptions.BaseCRMValidationError):
            utils.validate_contact_dict(operation, contact_dict)

        operation = utils.CREATE
        contact_dict = {
            'is_organization': False,
            'name': 'Acme Carburretors Inc.'
        }
        with self.assertRaises(exceptions.BaseCRMValidationError):
            utils.validate_contact_dict(operation, contact_dict)

        operation = utils.UPDATE
        contact_dict = {
            'is_organization': False,
            'name': 'Acme Carburretors Inc.'
        }
        with self.assertRaises(exceptions.BaseCRMValidationError):
            utils.validate_contact_dict(operation, contact_dict)

        operation = utils.CREATE
        contact_dict = {
            'is_organization': False,
            'first_name': 'Albert',
            'last_name': 'Einstein'
        }
        result = utils.validate_contact_dict(operation, contact_dict)
        self.assertTrue(result)

        operation = utils.UPDATE
        contact_dict = {
            'is_organization': False,
            'first_name': 'Albert',
            'last_name': 'Einstein'
        }
        with self.assertRaises(exceptions.BaseCRMValidationError):
            utils.validate_contact_dict(operation, contact_dict)

        operation = utils.CREATE
        contact_dict = {
            'id': 32
        }
        with self.assertRaises(exceptions.BaseCRMValidationError):
            utils.validate_contact_dict(operation, contact_dict)

        operation = utils.UPDATE
        contact_dict = {
            'id': 32
        }
        result = utils.validate_contact_dict(operation, contact_dict)
        self.assertTrue(result)

        operation = utils.UPDATE
        contact_dict = {
            'id': 32
        }
        skip_id = False
        suppress = True
        with self.assertRaises(NotImplementedError):
            utils.validate_contact_dict(operation, contact_dict, skip_id, suppress)

        operation = 'hello'
        contact_dict = {
            'id': 32
        }
        suppress = False
        with self.assertRaises(exceptions.BaseCRMValidationError):
            utils.validate_contact_dict(operation, contact_dict, skip_id, suppress)

        operation = 'GET'
        contact_dict = {
            'id': 32
        }
        suppress = False
        with self.assertRaises(exceptions.BaseCRMValidationError):
            utils.validate_contact_dict(operation, contact_dict, skip_id, suppress)

        operation = utils.UPDATE
        contact_dict = {
            'id': 32
        }
        suppress = 'hello'
        result = utils.validate_contact_dict(operation, contact_dict, skip_id, suppress)
        self.assertTrue(result)

        operation = utils.UPDATE
        contact_dict = {
            'id': 32
        }
        suppress = False
        result = utils.validate_contact_dict(operation, contact_dict, skip_id, suppress)
        self.assertTrue(result)

        operation = utils.UPDATE
        contact_dict = {}
        skip_id = True
        result = utils.validate_contact_dict(operation, contact_dict, skip_id, suppress)
        self.assertTrue(result)

    def test_validate_deal_dict(self):
        operation = utils.UPDATE
        deal_dict = {
            'id': 32
        }
        skip_id = False
        suppress = True
        with self.assertRaises(NotImplementedError):
            utils.validate_deal_dict(operation, deal_dict, skip_id, suppress)

        operation = utils.UPDATE
        deal_dict = {
            'id': 32
        }
        suppress = False
        result = utils.validate_deal_dict(operation, deal_dict, skip_id, suppress)
        self.assertTrue(result)

        operation = utils.UPDATE
        deal_dict = {
            'id': 32
        }
        suppress = 'hello'
        result = utils.validate_deal_dict(operation, deal_dict, skip_id, suppress)
        self.assertTrue(result)

        skip_id = True
        deal_dict = {}
        result = utils.validate_deal_dict(operation, deal_dict, skip_id, suppress)
        self.assertTrue(result)

        skip_id = False
        operation = utils.CREATE
        deal_dict = {
            'name': 'Manhattan Project',
            'contact_id': 23456,
            'custom_fields': {}
        }
        result = utils.validate_deal_dict(operation, deal_dict, skip_id, suppress)
        self.assertTrue(result)

        operation = utils.CREATE
        deal_dict = {
            'name': True,
            'contact_id': 'not really an id',
            'custom_fields': 'not_a_dict'
        }
        result = utils.validate_deal_dict(operation, deal_dict, skip_id, suppress)
        self.assertTrue(result)

        operation = utils.CREATE
        deal_dict = {
            'id': 32
        }
        with self.assertRaises(exceptions.BaseCRMValidationError):
            utils.validate_deal_dict(operation, deal_dict)

        skip_id = True
        operation = utils.CREATE
        deal_dict = {
            'id': 32
        }
        with self.assertRaises(exceptions.BaseCRMValidationError):
            utils.validate_deal_dict(operation, deal_dict)

        skip_id = False
        operation = utils.CREATE
        deal_dict = {
            'name': 'Manhattan Project',
        }
        with self.assertRaises(exceptions.BaseCRMValidationError):
            utils.validate_deal_dict(operation, deal_dict)

        operation = utils.CREATE
        deal_dict = {
            'name': 'Manhattan Project',
            'contact_id': 23456,
        }
        with self.assertRaises(exceptions.BaseCRMValidationError):
            utils.validate_deal_dict(operation, deal_dict)

        operation = utils.CREATE
        deal_dict = {
            'name': 'Manhattan Project',
            'custom_fields': 'not_a_dict'
        }
        with self.assertRaises(exceptions.BaseCRMValidationError):
            utils.validate_deal_dict(operation, deal_dict)

        operation = utils.UPDATE
        deal_dict = {
            'name': True,
            'contact_id': 'not really an id',
            'custom_fields': 'not_a_dict'
        }
        with self.assertRaises(exceptions.BaseCRMValidationError):
            utils.validate_deal_dict(operation, deal_dict)

        operation = utils.UPDATE
        deal_dict = {
            'id': True
        }
        result = utils.validate_deal_dict(operation, deal_dict)
        self.assertTrue(result)


class UtilityMethodTests(TestCase):

    @mock.patch('%s.utils.django_apps' % __name__)
    def test_instantiate_if_necessary(self, _django_apps):
        _app = mock.Mock()
        _app.pipeline = None
        _app.stages = None
        _app.users = None
        _django_apps.get_app_config.return_value = _app

        utils.instantiate_if_necessary()
        _django_apps.get_app_config.assert_called_once_with('django_basecrm')
        _app.instantiate_stages.assert_called_once()

        _django_apps.get_app_config.reset_mock()
        _app.instantiate_stages.reset_mock()

        _app.pipeline = 9877556
        _app.stages = None
        _app.users = None

        utils.instantiate_if_necessary()
        _django_apps.get_app_config.assert_called_once_with('django_basecrm')
        _app.instantiate_stages.assert_called_once()

        _django_apps.get_app_config.reset_mock()
        _app.instantiate_stages.reset_mock()

        _app.pipeline = 9877556
        _app.stages = ['one', 'two']
        _app.users = None

        utils.instantiate_if_necessary()
        _django_apps.get_app_config.assert_called_once_with('django_basecrm')
        _app.instantiate_stages.assert_called_once()

        _django_apps.get_app_config.reset_mock()
        _app.instantiate_stages.reset_mock()

        _app.pipeline = 9877556
        _app.stages = ['one', 'two']
        _app.users = ['a', 'b', 'c']

        utils.instantiate_if_necessary()
        _django_apps.get_app_config.assert_called_once_with('django_basecrm')
        self.assertFalse(_app.instantiate_stages.called)

    def test_count(self):
        response_dict = {}
        with self.assertRaises(exceptions.BaseCRMBadParameterFormat):
            utils.count(response_dict)

        response_dict = []
        with self.assertRaises(exceptions.BaseCRMBadParameterFormat):
            utils.count(response_dict)

        response_dict = ('hello', )
        with self.assertRaises(exceptions.BaseCRMBadParameterFormat):
            utils.count(response_dict)

        response_dict = {'name': 'hello'}
        with self.assertRaises(exceptions.BaseCRMBadParameterFormat):
            utils.count(response_dict)

        response_dict = {'name': 'hello', 'meta': []}
        result = utils.count(response_dict)
        self.assertEqual(result, 1)

        response_dict = {'name': 'hello', 'meta': 'world'}
        result = utils.count(response_dict)
        self.assertEqual(result, 1)

        response_dict = {'name': 'hello', 'meta': {'count': 98}}
        result = utils.count(response_dict)
        self.assertEqual(result, 98)

    def test_parse(self):
        response_dict = {}
        with self.assertRaises(exceptions.BaseCRMBadParameterFormat):
            utils.parse(response_dict)

        response_dict = []
        with self.assertRaises(exceptions.BaseCRMBadParameterFormat):
            utils.parse(response_dict)

        response_dict = ('hello', )
        with self.assertRaises(exceptions.BaseCRMBadParameterFormat):
            utils.parse(response_dict)

        response_dict = {'items': []}
        result = utils.parse(response_dict)
        self.assertEqual(result, [])

        response_dict = {'items': [{'name': 'hello'}]}
        with self.assertRaises(exceptions.BaseCRMBadParameterFormat):
            utils.parse(response_dict)

        response_dict = {'items': [{'data': {'name': 'hello'}}], 'meta': {'count': 98}}
        result = utils.parse(response_dict)
        self.assertEqual(result, [{'name': 'hello'}])

        response_dict = {'items': [{'data': {'name': 'hello'}}, {'data': {'name': 'world'}}], 'meta': {'count': 98}}
        result = utils.parse(response_dict)
        self.assertEqual(result, [{'name': 'hello'}, {'name': 'world'}])

        response_dict = {'name': 'hello'}
        with self.assertRaises(exceptions.BaseCRMBadParameterFormat):
            utils.parse(response_dict)

        response_dict = {'data': {'name': 'hello'}, 'meta': 'world'}
        result = utils.parse(response_dict)
        self.assertEqual(result, {'name': 'hello'})

    @mock.patch('%s.utils._request' % __name__)
    def test_request(self, _request):
        response = mock.Mock()
        response.status_code = 200
        response.json.return_value = {'items': [], 'meta': {}}
        _request.return_value = response


        # test all the pre-send validation

        action = 'GET'
        endpoint = 'contacts'
        get_params = None
        kwargs = {}

        with self.assertRaises(exceptions.BaseCRMBadParameterFormat):
            utils.request(action, endpoint, get_params, **kwargs)

        action = 'hello'
        with self.assertRaises(exceptions.BaseCRMBadParameterFormat):
            utils.request(action, endpoint, get_params, **kwargs)

        action = utils.RETRIEVE
        result = utils.request(action, endpoint, get_params, **kwargs)
        self.assertEqual(result, response.json.return_value)
        response.json.assert_called_once()

        response.json.reset_mock()
        action = utils.INFO
        result = utils.request(action, endpoint, get_params, **kwargs)
        self.assertEqual(result, response.json.return_value)
        response.json.assert_called_once()

        response.json.reset_mock()
        action = utils.DELETE
        result = utils.request(action, endpoint, get_params, **kwargs)
        self.assertEqual(result, response.json.return_value)
        response.json.assert_called_once()

        response.json.reset_mock()
        action = utils.CREATE
        with self.assertRaises(exceptions.BaseCRMBadParameterFormat):
            utils.request(action, endpoint, get_params, **kwargs)

        response.json.reset_mock()
        action = utils.UPDATE
        with self.assertRaises(exceptions.BaseCRMBadParameterFormat):
            utils.request(action, endpoint, get_params, **kwargs)

        kwargs['data'] = {}

        response.json.reset_mock()
        action = utils.CREATE
        result = utils.request(action, endpoint, get_params, **kwargs)
        self.assertEqual(result, response.json.return_value)
        response.json.assert_called_once()

        response.json.reset_mock()
        action = utils.UPDATE
        result = utils.request(action, endpoint, get_params, **kwargs)
        self.assertEqual(result, response.json.return_value)
        response.json.assert_called_once()

        # test the param logic

        _request.reset_mock()
        response.json.reset_mock()
        action = utils.UPDATE
        result = utils.request(action, endpoint, get_params, **kwargs)
        self.assertEqual(result, response.json.return_value)
        _request.assert_called_once_with(
            utils.VERBS[action],
            endpoint,
            get_params,
            json={'data': {}}
        )
        response.json.assert_called_once()

        _request.reset_mock()
        response.json.reset_mock()
        action = utils.CREATE
        kwargs = {
            'data': {'serialized_id': 98},
            'headers': {'Accept': 'text/plain'},
            'other_param_for_requests_lib': True
        }
        result = utils.request(action, endpoint, get_params, **kwargs)
        self.assertEqual(result, response.json.return_value)
        _request.assert_called_once_with(
            utils.VERBS[action],
            endpoint,
            get_params,
            json={'data': {'serialized_id': 98}},
            headers={'Accept': 'text/plain'},
            other_param_for_requests_lib=True
        )
        response.json.assert_called_once()

        response.json.return_value = {'errors': [{'error': {'details': 'test error message'}}]}
        response.status_code = 401

        _request.reset_mock()
        response.json.reset_mock()
        action = utils.RETRIEVE
        kwargs = {}
        with self.assertRaises(exceptions.BaseCRMAPIUnauthorized):
            utils.request(action, endpoint, get_params, **kwargs)

        response.status_code = 403

        _request.reset_mock()
        response.json.reset_mock()
        action = utils.RETRIEVE
        kwargs = {}
        with self.assertRaises(Exception):
            utils.request(action, endpoint, get_params, **kwargs)

        response.status_code = 404

        _request.reset_mock()
        response.json.reset_mock()
        action = utils.RETRIEVE
        kwargs = {}
        with self.assertRaises(Exception):
            utils.request(action, endpoint, get_params, **kwargs)

        response.status_code = 404
        get_params = {'id': 99}

        _request.reset_mock()
        response.json.reset_mock()
        action = utils.RETRIEVE
        kwargs = {}
        with self.assertRaises(exceptions.BaseCRMNoResult):
            utils.request(action, endpoint, get_params, **kwargs)

        response.status_code = 422
        get_params = None

        _request.reset_mock()
        response.json.reset_mock()
        action = utils.RETRIEVE
        kwargs = {}
        with self.assertRaises(exceptions.BaseCRMValidationError):
            utils.request(action, endpoint, get_params, **kwargs)


class HelperMethodTests(TestCase):

    @mock.patch('%s.utils.request' % __name__)
    @mock.patch('%s.utils.parse' % __name__)
    def test_get_contacts(self, parse, request):
        request.return_value = {
            'items':[{'id':23, 'name':'hello'},{'id':99, 'name':'world'}],
            'meta': {'count': 2}
        }
        parse.return_value = [{'id':23, 'name':'hello'},{'id':99, 'name':'world'}]

        result = helpers.get_contacts()
        self.assertEqual(result, parse.return_value)
        request.assert_called_once_with(utils.RETRIEVE, 'contacts', {})
        parse.assert_called_once_with(request.return_value)

        request.reset_mock()
        parse.reset_mock()
        result = helpers.get_contacts(id=456, hello='world')
        self.assertEqual(result, parse.return_value)
        request.assert_called_once_with(utils.RETRIEVE, 'contacts', {'id': 456, 'hello': 'world'})
        parse.assert_called_once_with(request.return_value)

    @mock.patch('%s.utils.validate_contact_dict' % __name__)
    @mock.patch('%s.utils.request' % __name__)
    @mock.patch('%s.utils.parse' % __name__)
    def test_create_contact(self, parse, request, validate):
        request.return_value = {
            'data':{'id':23, 'name':'M deLaurentiis'},
            'meta': {'count': 1}
        }
        parse.return_value = {'id':23, 'name':'M deLaurentiis'}
        validate.return_value = True
        data = {'id': 999, 'name': 'M deLaurentiis'}

        result = helpers.create_contact(data)
        self.assertEqual(result, parse.return_value)
        validate.assert_called_once_with(utils.CREATE, data)
        request.assert_called_once_with(utils.CREATE, 'contacts', None, data=data)
        parse.assert_called_once_with(request.return_value)

        request.reset_mock()
        parse.reset_mock()
        validate.reset_mock()
        validate.side_effect = exceptions.BaseCRMValidationError()
        with self.assertRaises(exceptions.BaseCRMValidationError):
            helpers.create_contact(data)

        request.reset_mock()
        parse.reset_mock()
        validate.reset_mock()
        validate.side_effect = None
        validate.return_value = False
        result = helpers.create_contact(data)
        validate.assert_called_once_with(utils.CREATE, data)
        self.assertFalse(request.called)
        self.assertFalse(parse.called)

    @mock.patch('%s.utils.validate_contact_dict' % __name__)
    @mock.patch('%s.utils.request' % __name__)
    @mock.patch('%s.utils.parse' % __name__)
    def test_update_contact(self, parse, request, validate):
        request.return_value = {
            'data':{'id':23, 'name':'G deLaurentiis'},
            'meta': {'count': 1}
        }
        parse.return_value = {'id':23, 'name':'G deLaurentiis'}
        validate.return_value = True
        data = {'name': 'G deLaurentiis'}
        id = 33

        result = helpers.update_contact(id, data)
        self.assertEqual(result, parse.return_value)
        validate.assert_called_once_with(utils.UPDATE, data, skip_id=True)
        request.assert_called_once_with(utils.UPDATE, 'contacts', {'id': id}, data=data)
        parse.assert_called_once_with(request.return_value)

        request.reset_mock()
        parse.reset_mock()
        validate.reset_mock()
        validate.side_effect = exceptions.BaseCRMValidationError()
        with self.assertRaises(exceptions.BaseCRMValidationError):
            helpers.update_contact(id, data)

        request.reset_mock()
        parse.reset_mock()
        validate.reset_mock()
        validate.side_effect = None
        validate.return_value = False
        result = helpers.update_contact(id, data)
        validate.assert_called_once_with(utils.UPDATE, data, skip_id=True)
        self.assertFalse(request.called)
        self.assertFalse(parse.called)

    @mock.patch('%s.utils.request' % __name__)
    @mock.patch('%s.utils.parse' % __name__)
    def test_get_deals(self, parse, request):
        request.return_value = {
            'items':[{'id':23, 'name':'hello'},{'id':99, 'name':'world'}],
            'meta': {'count': 2}
        }
        parse.return_value = [{'id':23, 'name':'hello'},{'id':99, 'name':'world'}]

        result = helpers.get_deals()
        self.assertEqual(result, parse.return_value)
        request.assert_called_once_with(utils.RETRIEVE, 'deals', {})
        parse.assert_called_once_with(request.return_value)

        request.reset_mock()
        parse.reset_mock()
        result = helpers.get_deals(id=456, hello='world')
        self.assertEqual(result, parse.return_value)
        request.assert_called_once_with(utils.RETRIEVE, 'deals', {'id': 456, 'hello': 'world'})
        parse.assert_called_once_with(request.return_value)

    @mock.patch('%s.utils.validate_deal_dict' % __name__)
    @mock.patch('%s.utils.request' % __name__)
    @mock.patch('%s.utils.parse' % __name__)
    def test_create_deal(self, parse, request, validate):
        request.return_value = {
            'data':{'id':23, 'name':'Создание советской атомной бомбы'},
            'meta': {'count': 1}
        }
        parse.return_value = {'id':23, 'name':'Создание советской атомной бомбы'}
        validate.return_value = True
        data = {'id': 999, 'name': 'Создание советской атомной бомбы'}

        result = helpers.create_deal(data)
        self.assertEqual(result, parse.return_value)
        validate.assert_called_once_with(utils.CREATE, data)
        request.assert_called_once_with(utils.CREATE, 'deals', None, data=data)
        parse.assert_called_once_with(request.return_value)

        request.reset_mock()
        parse.reset_mock()
        validate.reset_mock()
        validate.side_effect = exceptions.BaseCRMValidationError()
        with self.assertRaises(exceptions.BaseCRMValidationError):
            helpers.create_deal(data)

        request.reset_mock()
        parse.reset_mock()
        validate.reset_mock()
        validate.side_effect = None
        validate.return_value = False
        result = helpers.create_deal(data)
        validate.assert_called_once_with(utils.CREATE, data)
        self.assertFalse(request.called)
        self.assertFalse(parse.called)

    @mock.patch('%s.utils.validate_deal_dict' % __name__)
    @mock.patch('%s.utils.request' % __name__)
    @mock.patch('%s.utils.parse' % __name__)
    def test_update_deal(self, parse, request, validate):
        request.return_value = {
            'data':{'id':23, 'name':'Manhattan Project'},
            'meta': {'count': 1}
        }
        parse.return_value = {'id':23, 'name':'Manhattan Project'}
        validate.return_value = True
        data = {'id': 999, 'name': 'Manhattan Project'}
        id = 23

        result = helpers.update_deal(id, data)
        self.assertEqual(result, parse.return_value)
        validate.assert_called_once_with(utils.UPDATE, data, skip_id=True)
        request.assert_called_once_with(utils.UPDATE, 'deals', {'id': id}, data=data)
        parse.assert_called_once_with(request.return_value)

        request.reset_mock()
        parse.reset_mock()
        validate.reset_mock()
        validate.side_effect = exceptions.BaseCRMValidationError()
        with self.assertRaises(exceptions.BaseCRMValidationError):
            helpers.update_deal(id, data)

        request.reset_mock()
        parse.reset_mock()
        validate.reset_mock()
        validate.side_effect = None
        validate.return_value = False
        result = helpers.update_deal(id, data)
        validate.assert_called_once_with(utils.UPDATE, data, skip_id=True)
        self.assertFalse(request.called)
        self.assertFalse(parse.called)

    @mock.patch('%s.utils.request' % __name__)
    @mock.patch('%s.utils.parse' % __name__)
    def test_get_notes(self, parse, request):
        request.return_value = {
            'items':[{'id':23, 'content':'hello'},{'id':99, 'content':'world'}],
            'meta': {'count': 2}
        }
        parse.return_value = [{'id':23, 'content':'hello'},{'id':99, 'content':'world'}]
        resource_type = 'contact'
        resource_id = 55

        result = helpers.get_notes()
        self.assertEqual(result, parse.return_value)
        request.assert_called_once_with(utils.RETRIEVE, 'notes', {})
        parse.assert_called_once_with(request.return_value)

        request.reset_mock()
        parse.reset_mock()
        result = helpers.get_notes(resource_type)
        self.assertEqual(result, parse.return_value)
        request.assert_called_once_with(utils.RETRIEVE, 'notes', {'resource_type': 'contact'})
        parse.assert_called_once_with(request.return_value)

        request.reset_mock()
        parse.reset_mock()
        resource_type = 'lead'
        result = helpers.get_notes(resource_type, resource_id)
        self.assertEqual(result, parse.return_value)
        request.assert_called_once_with(utils.RETRIEVE, 'notes', {'resource_type': 'lead', 'resource_id': 55})
        parse.assert_called_once_with(request.return_value)

        request.reset_mock()
        parse.reset_mock()
        resource_type = 'deal'
        result = helpers.get_notes(resource_type, resource_id, page=5)
        self.assertEqual(result, parse.return_value)
        request.assert_called_once_with(utils.RETRIEVE, 'notes', {'resource_type': 'deal', 'resource_id': 55, 'page': 5})
        parse.assert_called_once_with(request.return_value)

        with self.assertRaises(exceptions.BaseCRMValidationError):
            helpers.get_notes(resource_id)

        resource_type = 'contacts'
        with self.assertRaises(exceptions.BaseCRMValidationError):
            helpers.get_notes(resource_type)

        resource_type = 56
        with self.assertRaises(exceptions.BaseCRMValidationError):
            helpers.get_notes(resource_type)

    @mock.patch('%s.utils.instantiate_if_necessary' % __name__)
    @mock.patch('%s.helpers.django_apps' % __name__)
    def test_get_pipelines(self, _apps, instantiate):
        _app_conf = mock.Mock()
        _app_conf.pipeline = {'id': 6456, 'name': 'default'}
        _apps.get_app_config.return_value = _app_conf

        result = helpers.get_pipelines()
        self.assertEqual(result, _app_conf.pipeline)
        _apps.get_app_config.assert_called_once_with('django_basecrm')
        instantiate.assert_called_once()

    @mock.patch('%s.helpers.get_stages_from_api' % __name__)
    @mock.patch('%s.helpers.base_settings' % __name__)
    @mock.patch('%s.apps.BaseCRMConfig.instantiate_stages' % __name__)
    @mock.patch('%s.helpers.django_apps' % __name__)
    def test_get_stages(self, _apps, instantiate, settings, get_from_api):
        settings.BASECRM_CACHE_STAGES = True
        _app_conf = mock.Mock()
        _app_conf.stages = [{'id': 6456, 'name': 'new'}, {'id': 6577, 'name': 'updated'}]
        get_from_api.return_value = [{'id': 6993, 'name': 'won'}, {'id': 7004, 'name': 'lost'}]
        _apps.get_app_config.return_value = _app_conf

        result = helpers.get_stages()
        self.assertEqual(result, _app_conf.stages)
        _apps.get_app_config.assert_called_once_with('django_basecrm')
        instantiate.assert_called_once()
        self.assertFalse(get_from_api.called)

        settings.BASECRM_CACHE_STAGES = False
        instantiate.reset_mock()
        _apps.get_app_config.reset_mock()

        result = helpers.get_stages()
        self.assertEqual(result, get_from_api.return_value)
        self.assertFalse(_apps.get_app_config.called)
        self.assertFalse(instantiate.called)
        get_from_api.assert_called_once()

    @mock.patch('%s.helpers.get_users_from_api' % __name__)
    @mock.patch('%s.helpers.base_settings' % __name__)
    @mock.patch('%s.apps.BaseCRMConfig.instantiate_users' % __name__)
    @mock.patch('%s.helpers.django_apps' % __name__)
    def test_get_users(self, _apps, instantiate, settings, get_from_api):
        settings.BASECRM_CACHE_USERS = True
        _app_conf = mock.Mock()
        _app_conf.users = [{'id': 6456, 'name': 'Albert'}, {'id': 6577, 'name': 'Bertie'}]
        get_from_api.return_value = [{'id': 6993, 'name': 'Charlie'}, {'id': 7004, 'name': 'Davie'}]
        _apps.get_app_config.return_value = _app_conf

        result = helpers.get_users()
        self.assertEqual(result, _app_conf.users)
        _apps.get_app_config.assert_called_once_with('django_basecrm')
        instantiate.assert_called_once()
        self.assertFalse(get_from_api.called)

        settings.BASECRM_CACHE_USERS = False
        instantiate.reset_mock()
        _apps.get_app_config.reset_mock()

        result = helpers.get_users()
        self.assertEqual(result, get_from_api.return_value)
        self.assertFalse(_apps.get_app_config.called)
        self.assertFalse(instantiate.called)
        get_from_api.assert_called_once()

    @mock.patch('%s.helpers.get_stages' % __name__)
    def test_get_stage_ids(self, get_stages):
        get_stages.return_value = [{'id': 8888, 'name': 'New'}, {'id': 9999, 'name': 'In Progress'}]

        result = helpers.get_stage_ids()
        self.assertEqual(result, [8888, 9999])
        get_stages.assert_called_once()

    @mock.patch('%s.helpers.get_users' % __name__)
    def test_get_user_ids(self, get_users):
        get_users.return_value = [{'id': 8, 'name': 'Albert'}, {'id': 9, 'name': 'Berties'}]

        result = helpers.get_user_ids()
        self.assertEqual(result, [8, 9])
        get_users.assert_called_once()

    @mock.patch('%s.utils.request' % __name__)
    @mock.patch('%s.utils.parse' % __name__)
    @mock.patch('%s.utils.count' % __name__)
    def test_get_pipelines_from_api(self, count, parse, request):
        request.return_value = {
            'items':[{'id':23, 'name':'hello'},{'id':99, 'name':'world'}],
            'meta': {'count': 2}
        }
        parse.return_value = [{'id':23, 'name':'hello'},{'id':99, 'name':'world'}]
        count.return_value = 2

        with self.assertRaises(NotImplementedError):
            helpers.get_pipelines_from_api()
            request.assert_called_once_with(utils.RETRIEVE, 'pipelines', {})
            count.assert_called_once_with(request.return_value)

        request.reset_mock()
        count.reset_mock()
        parse.reset_mock()
        count.return_value = 1
        result = helpers.get_pipelines_from_api()
        self.assertEqual(result, parse.return_value)
        request.assert_called_once_with(utils.RETRIEVE, 'pipelines', {})
        count.assert_called_once_with(request.return_value)
        parse.assert_called_once_with(request.return_value)

        request.reset_mock()
        count.reset_mock()
        parse.reset_mock()
        result = helpers.get_pipelines_from_api(id=456, hello='world')
        self.assertEqual(result, parse.return_value)
        request.assert_called_once_with(utils.RETRIEVE, 'pipelines', {'id': 456, 'hello': 'world'})
        count.assert_called_once_with(request.return_value)
        parse.assert_called_once_with(request.return_value)

    @mock.patch('%s.utils.request' % __name__)
    @mock.patch('%s.utils.parse' % __name__)
    def test_get_stages_from_api(self, parse, request):
        request.return_value = {
            'items':[{'id':23, 'name':'new'},{'id':99, 'name':'called in'}],
            'meta': {'count': 2}
        }
        parse.return_value = [{'id':23, 'name':'new'},{'id':99, 'name':'called in'}]

        result = helpers.get_stages_from_api()
        self.assertEqual(result, parse.return_value)
        request.assert_called_once_with(utils.RETRIEVE, 'stages', {})
        parse.assert_called_once_with(request.return_value)

        request.reset_mock()
        parse.reset_mock()
        result = helpers.get_stages_from_api(id=456, hello='world')
        self.assertEqual(result, parse.return_value)
        request.assert_called_once_with(utils.RETRIEVE, 'stages', {'id': 456, 'hello': 'world'})
        parse.assert_called_once_with(request.return_value)

    @mock.patch('%s.utils.request' % __name__)
    @mock.patch('%s.utils.parse' % __name__)
    def test_get_users_from_api(self, parse, request):
        request.return_value = {
            'items':[{'id':23, 'name':'Albert'},{'id':99, 'name':'Bertie'}],
            'meta': {'count': 2}
        }
        parse.return_value = [{'id':23, 'name':'Albert'},{'id':99, 'name':'Bertie'}]

        result = helpers.get_users_from_api()
        self.assertEqual(result, parse.return_value)
        request.assert_called_once_with(utils.RETRIEVE, 'users', {})
        parse.assert_called_once_with(request.return_value)

        request.reset_mock()
        parse.reset_mock()
        result = helpers.get_users_from_api(id=456, hello='world')
        self.assertEqual(result, parse.return_value)
        request.assert_called_once_with(utils.RETRIEVE, 'users', {'id': 456, 'hello': 'world'})
        parse.assert_called_once_with(request.return_value)


class SerializerTests(TestCase):

    def setUp(self):
        self.instance = mock.Mock(spec=ModelBase)
        class ExampleAbstractSerializer(serializers.AbstractModelSerializer):
            base_fields = []
            class Meta:
                model = self.instance.__class__
                fields = [
                    'dummy',
                    'another'
                ]
        class ExampleContactSerializer(serializers.ContactModelSerializer):
            class Meta:
                model = self.instance.__class__
        class ExampleDealSerializer(serializers.DealModelSerializer):
            class Meta:
                model = self.instance.__class__
        self.serialized_abstract_object = ExampleAbstractSerializer(self.instance)
        self.serialized_contact = ExampleContactSerializer(self.instance)
        self.serialized_deal = ExampleDealSerializer(self.instance)
        self.ExampleContactSerializer = ExampleContactSerializer
        self.ExampleDealSerializer = ExampleDealSerializer

    @mock.patch('%s.serializers.AbstractModelSerializer._self_assign_values' % __name__)
    def test_init(self, assign):
        self.assertTrue(hasattr(self.serialized_abstract_object, 'instance'))
        self.assertTrue(hasattr(self.serialized_abstract_object, 'base_fields'))
        self.assertTrue(hasattr(self.serialized_abstract_object, 'read_only_fields'))
        self.assertTrue(hasattr(self.serialized_abstract_object.Meta, 'model'))
        self.assertTrue(hasattr(self.serialized_abstract_object.Meta, 'fields'))

        self.assertEqual(self.serialized_abstract_object.instance, self.instance)
        self.assertEqual(self.serialized_abstract_object.base_fields, [])
        self.assertEqual(self.serialized_abstract_object.read_only_fields, None)
        self.assertEqual(self.serialized_abstract_object.Meta.model, self.instance.__class__)
        self.assertEqual(self.serialized_abstract_object.Meta.fields, ['dummy', 'another'])

        assign.assert_called_once()

    def test_to_dict(self):
        self.instance.name = 'Acme Enterprises Inc.'
        class ExampleOrganizationSerializer(self.ExampleContactSerializer):
            is_organization = False

        serialized_organization = ExampleOrganizationSerializer(self.instance)
        result = serialized_organization.to_dict()

        self.assertEqual(result, {
            'is_organization': False,
            'name': 'Acme Enterprises Inc.'
        })

    @mock.patch('%s.serializers.AbstractModelSerializer._get_field_list' % __name__)
    @mock.patch('%s.serializers.AbstractModelSerializer._get_value' % __name__)
    def test_self_assign_values(self, get_value, get_fields):
        get_fields.return_value = ['id', 'name']
        get_value.side_effect = ['Albert Einstein']
        self.assertEqual(self.ExampleContactSerializer.read_only_fields, ['id'])
        serialized_contact = self.ExampleContactSerializer(self.instance)

        self.assertFalse(hasattr(serialized_contact, 'id'))
        self.assertTrue(hasattr(serialized_contact, 'name'))
        self.assertEqual(serialized_contact.name, 'Albert Einstein')
        get_fields.assert_called_once()
        get_value.assert_called_once_with('name')

        get_fields.reset_mock()
        get_value.reset_mock()
        get_fields.return_value = ['id', 'name', 'is_organization', 'is_deceased']
        get_value.side_effect = ['John von Neumann', False, None]
        serialized_contact = self.ExampleContactSerializer(self.instance)

        self.assertFalse(hasattr(serialized_contact, 'id'))
        self.assertTrue(hasattr(serialized_contact, 'name'))
        self.assertEqual(serialized_contact.name, 'John von Neumann')
        self.assertTrue(hasattr(serialized_contact, 'is_organization'))
        self.assertEqual(serialized_contact.is_organization, False)
        self.assertFalse(hasattr(serialized_contact, 'is_deceased'))
        get_fields.assert_called_once()
        self.assertEqual(get_value.call_count, 3)

    def test_get_field_list(self):
        class BaseFieldAbstractSerializer(serializers.AbstractModelSerializer):
            base_fields = [
                'test',
                'another',
                'third'
            ]
            class Meta:
                model = self.instance.__class__
        class MetaFieldAbstractSerializer(serializers.AbstractModelSerializer):
            base_fields = [
                'test',
                'another',
                'third'
            ]
            class Meta:
                model = self.instance.__class__
                fields = [
                    'meta_test',
                    'test_meta',
                    'another_test'
                ]

        base_serialized = BaseFieldAbstractSerializer(self.instance)
        meta_serialized = MetaFieldAbstractSerializer(self.instance)
        results = base_serialized._get_field_list()
        self.assertEqual(
            results,
            [
                'test',
                'another',
                'third'
            ]
        )
        results = meta_serialized._get_field_list()
        self.assertEqual(
            results,
            [
                'meta_test',
                'test_meta',
                'another_test'
            ]
        )

    def test_get_value(self):
        class TestSerializer(self.ExampleContactSerializer):
            is_organization = False
            name = 'Test'
            phone = 'phone'
            email = 'email'
            description = 'desc'

            def get_hot(self, obj):
                return obj.am_i_hot_today

            class Meta:
                model = self.instance.__class__
                fields = [
                    'is_organization',
                    'name',
                    'phone',
                    'email',
                    'hot',
                    'description',
                    'custom_field',
                    'extra_field'
                ]

        self.instance.name = "Not Test"
        self.instance.desc = "Test description"
        self.instance.email = "test@domain.com"
        self.instance.am_i_hot_today = True
        self.instance.custom_field = 'test value'
        self.instance.phone = mock.Mock()
        self.instance.phone.return_value = "0208555888"
        self.instance.phone.__class__ = types.MethodType
        serialized = TestSerializer(self.instance)

        # Fields that don't exist
        with self.assertRaises(TypeError):
            serialized._get_value(87)

        result = serialized._get_value('not_existing')
        self.assertEqual(result, None)

        # fields that are configured but not on instance
        result = serialized._get_value('extra_field')
        self.assertEqual(result, None)

        # Instance attributes / properties
        result = serialized._get_value('email')
        self.assertEqual(result, "test@domain.com")

        # ... and one with a non-identically named field mapping
        result = serialized._get_value('description')
        self.assertEqual(result, "Test description")

        # ... and one that isn't specified
        result = serialized._get_value('custom_field')
        self.assertEqual(result, "test value")

        # Instance methods
        result = serialized._get_value('phone')
        self.assertEqual(result, "0208555888")
        self.instance.phone.assert_called_once()

        # Class level values
        result = serialized._get_value('is_organization')
        self.assertFalse(result)

        # ... overridden by instance values
        result = serialized._get_value('name')
        self.assertEqual(result, "Not Test")

        # Class level get_ methods
        result = serialized._get_value('hot')
        self.assertTrue(result)

    def test_contact_fields(self):
        self.assertTrue(hasattr(self.serialized_contact, 'instance'))
        self.assertTrue(hasattr(self.serialized_contact, 'base_fields'))
        self.assertTrue(hasattr(self.serialized_contact, 'read_only_fields'))
        self.assertTrue(hasattr(self.serialized_contact.Meta, 'model'))

        self.assertEqual(self.serialized_contact.instance, self.instance)
        self.assertEqual(self.serialized_contact.base_fields, [
            'id',
            'owner_id',
            'is_organization',
            'contact_id',
            'name',
            'first_name',
            'last_name',
            'customer_status',
            'prospect_status',
            'title',
            'description',
            'industry',
            'website',
            'email',
            'phone',
            'mobile',
            'fax',
            'twitter',
            'facebook',
            'linkedin',
            'skype',
            'address',
            'tags',
            'custom_fields',
        ])
        self.assertEqual(self.serialized_contact.read_only_fields, ['id'])
        self.assertEqual(self.serialized_contact.Meta.model, self.instance.__class__)

    def test_deal_fields(self):
        self.assertTrue(hasattr(self.serialized_deal, 'instance'))
        self.assertTrue(hasattr(self.serialized_deal, 'base_fields'))
        self.assertTrue(hasattr(self.serialized_deal, 'read_only_fields'))
        self.assertTrue(hasattr(self.serialized_deal.Meta, 'model'))

        self.assertEqual(self.serialized_deal.instance, self.instance)
        self.assertEqual(self.serialized_deal.base_fields, [
            'id',
            'owner_id',
            'name',
            'value',
            'currency',
            'hot',
            'stage_id',
            'source_id',
            'loss_reason_id',
            'dropbox_email',
            'contact_id',
            'organization_id',
            'tags',
            'custom_fields',
        ])
        self.assertEqual(self.serialized_deal.read_only_fields, ['id', 'organization_id'])
        self.assertEqual(self.serialized_deal.Meta.model, self.instance.__class__)

    def test_validate_field(self):
        result = self.serialized_abstract_object._validate_field('any', 'val')
        self.assertTrue(result)

    @mock.patch('%s.serializers.base_helpers' % __name__)
    def test_contact_validate_field(self, helpers):
        helpers.get_user_ids.return_value = [111, 222, 333, 444]

        result = self.serialized_contact._validate_field('any', True)
        self.assertTrue(result)

        result = self.serialized_contact._validate_field('owner_id', None)
        self.assertTrue(result)

        result = self.serialized_contact._validate_field('owner_id', True)
        self.assertFalse(result)

        result = self.serialized_contact._validate_field('owner_id', 777)
        self.assertFalse(result)

        result = self.serialized_contact._validate_field('owner_id', 444)
        self.assertTrue(result)

    @mock.patch('%s.serializers.base_helpers' % __name__)
    def test_deal_validate_field(self, helpers):
        helpers.get_user_ids.return_value = [111, 222, 333, 444]
        helpers.get_stage_ids.return_value = [666, 777, 888, 999]

        result = self.serialized_deal._validate_field('any', True)
        self.assertTrue(result)

        result = self.serialized_deal._validate_field('owner_id', None)
        self.assertTrue(result)

        result = self.serialized_deal._validate_field('owner_id', True)
        self.assertFalse(result)

        result = self.serialized_deal._validate_field('owner_id', 777)
        self.assertFalse(result)

        result = self.serialized_deal._validate_field('owner_id', 444)
        self.assertTrue(result)

        result = self.serialized_deal._validate_field('owner_id', None)
        self.assertTrue(result)

        result = self.serialized_deal._validate_field('stage_id', True)
        self.assertFalse(result)

        result = self.serialized_deal._validate_field('stage_id', 444)
        self.assertFalse(result)

        result = self.serialized_deal._validate_field('stage_id', 777)
        self.assertTrue(result)


class AppMethodTests(TestCase):

    def setUp(self):
        self.base_app = django_apps.get_app_config('django_basecrm')
        self.base_app.users = None
        self.base_app.stages = None
        self.base_app.pipeline = None

    @mock.patch('%s.apps.base_settings' % __name__)
    @mock.patch('%s.apps.BaseCRMConfig.instantiate_objects' % __name__)
    def test_ready(self, instantiate_objects, settings):
        settings.BASECRM_CACHE_AT_STARTUP = False
        result = self.base_app.ready()
        self.assertFalse(instantiate_objects.called)

        settings.BASECRM_CACHE_AT_STARTUP = True
        result = self.base_app.ready()
        instantiate_objects.assert_called_once()

    @mock.patch('%s.apps.get_pipelines_from_api' % __name__)
    def test_instantiate_pipeline(self, get_pipelines):
        get_pipelines.return_value = [{'id': 'one'}, {'id': 'two'}, {'id': 'three'}]
        self.assertEqual(self.base_app.pipeline, None)

        self.base_app.instantiate_pipeline()
        self.assertEqual(self.base_app.pipeline, {'id': 'one'})
        get_pipelines.assert_called_once()

        get_pipelines.return_value = {'id': 'two'}

        # get from cache
        self.base_app.instantiate_pipeline()
        self.assertEqual(self.base_app.pipeline, {'id': 'one'})
        self.assertEqual(get_pipelines.call_count, 1)

        # force get from API
        self.base_app.instantiate_pipeline(True)
        self.assertEqual(self.base_app.pipeline, {'id': 'two'})
        self.assertEqual(get_pipelines.call_count, 2)

        get_pipelines.return_value = [{'id': 'four'}, {'id': 'three'}]

        # confim only one pipeline kept
        self.base_app.instantiate_pipeline(True)
        self.assertEqual(self.base_app.pipeline, {'id': 'four'})

    @mock.patch('%s.apps.get_stages_from_api' % __name__)
    def test_instantiate_stages(self, get_stages):
        get_stages.return_value = ['A', 'B', 'C']
        self.assertEqual(self.base_app.stages, None)

        self.base_app.instantiate_stages()
        self.assertEqual(self.base_app.stages, ['A', 'B', 'C'])
        get_stages.assert_called_once()

        get_stages.return_value = ['D', 'E', 'F']

        # get from cache
        self.base_app.instantiate_stages()
        self.assertEqual(self.base_app.stages, ['A', 'B', 'C'])
        self.assertEqual(get_stages.call_count, 1)

        # force get from API
        self.base_app.instantiate_stages(True)
        self.assertEqual(self.base_app.stages, ['D', 'E', 'F'])
        self.assertEqual(get_stages.call_count, 2)

        get_stages.reset_mock()
        self.base_app.stages = None
        self.base_app.pipeline = {'id': 99999}
        self.base_app.instantiate_stages(True)
        get_stages.assert_called_once_with(pipeline_id=99999)

    @mock.patch('%s.apps.get_users_from_api' % __name__)
    def test_instantiate_users(self, get_users):
        get_users.return_value = [1, 2, 3]
        self.assertEqual(self.base_app.users, None)

        self.base_app.instantiate_users()
        self.assertEqual(self.base_app.users, [1, 2, 3])
        get_users.assert_called_once()

        get_users.return_value = [1, 55, 77]

        # get from cache
        self.base_app.instantiate_users()
        self.assertEqual(self.base_app.users, [1, 2, 3])
        self.assertEqual(get_users.call_count, 1)

        # force get from API
        self.base_app.instantiate_users(True)
        self.assertEqual(self.base_app.users, [1, 55, 77])
        self.assertEqual(get_users.call_count, 2)

    @mock.patch('%s.apps.BaseCRMConfig.instantiate_stages' % __name__)
    @mock.patch('%s.apps.BaseCRMConfig.instantiate_pipeline' % __name__)
    @mock.patch('%s.apps.BaseCRMConfig.instantiate_users' % __name__)
    def test_instantiate_objects(self, get_users, get_pipelines, get_stages):
        self.base_app.instantiate_objects()
        get_pipelines.assert_called_once()
        get_stages.assert_called_once()
        get_users.assert_called_once()

        get_pipelines.reset_mock()
        get_stages.reset_mock()
        get_users.reset_mock()

        self.base_app.instantiate_objects(True)
        get_pipelines.assert_called_once_with(True)
        get_stages.assert_called_once_with(True)
        get_users.assert_called_once_with(True)



# -*- coding: utf-8 -*-
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

class BaseCRMAPIUnauthorized(Exception):
    default_detail = _(u'Required access token is missing, malformed, expired, or invalid.')

    def __init__(self, detail=None):
        if detail is not None:
            self.detail = force_text(detail)
        else:
            self.detail = force_text(self.default_detail)

    def __str__(self):
        return self.detail

class BaseCRMBadParameterFormat(Exception):
    default_detail = _(
        u'Parameter format was not as expected: a full dict from the requests.json() '
        u'method applied to the full BaseCRM API response'
    )

    def __init__(self, detail=None):
        if detail is not None:
            self.detail = force_text(detail)
        else:
            self.detail = force_text(self.default_detail)

    def __str__(self):
        return self.detail

class BaseCRMValidationError(Exception):
    default_detail = _(
        u'Parameters fail BaseCRM API requirements'
    )

    def __init__(self, detail=None):
        if detail is not None:
            self.detail = force_text(detail)
        else:
            self.detail = force_text(self.default_detail)

    def __str__(self):
        return self.detail


class BaseCRMConfigurationError(Exception):
    default_detail = _(
        u'Configuration is not valid'
    )

    def __init__(self, detail=None):
        if detail is not None:
            self.detail = force_text(detail)
        else:
            self.detail = force_text(self.default_detail)

    def __str__(self):
        return self.detail


class BaseCRMUnexpectedResult(Exception):
    default_detail = _(
        u'Expected a different result format'
    )

    def __init__(self, detail=None):
        if detail is not None:
            self.detail = force_text(detail)
        else:
            self.detail = force_text(self.default_detail)

    def __str__(self):
        return self.detail


class BaseCRMNoResult(Exception):
    default_detail = _(
        u'No results exist for the given query'
    )

    def __init__(self, detail=None):
        if detail is not None:
            self.detail = force_text(detail)
        else:
            self.detail = force_text(self.default_detail)

    def __str__(self):
        return self.detail




# -*- coding: utf-8 -*-
import types

from django.apps import apps
from django.db.models.base import ModelBase
from django.db.models.loading import get_model

from . import (
    utils as base_api,
    exceptions as base_exceptions,
    helpers as base_helpers
)


class AbstractModelSerializer(object):
    """
    A helper for defining which attributes and properties of Django ORM models map to which BaseCRM
    fields. This is extended and used by concrete serializers below.
    """
    instance = None
    base_fields = None
    read_only_fields = None

    def __init__(self, instance, *args, **kwargs):
        """
        Checks this (extended) class has set the basics correctly, and triggers the value assignment
        """
        if self.Meta.model is None:
            raise base_exceptions.BaseCRMConfigurationError(
                "ModelSerializer must be defined with the relevant model on the Meta inner class"
            )
        if isinstance(self.Meta.model, basestring):
            try:
                self.Meta.model = get_model ( *self.Meta.model.split('.',1) )
            except:
                raise base_exceptions.BaseCRMConfigurationError(
                    "The model on the Meta inner class could not be derived from the string given"
                )
        if not isinstance(instance, self.Meta.model):
            raise base_exceptions.BaseCRMConfigurationError(
                "Initialise serializer with an instance of type model (as defined in Meta class)"
            )
        self.instance = instance
        self._self_assign_values()

    def to_dict(self):
        """
        Creates a dict of the self-set values
        """
        output = {}
        for f in self._get_field_list():
            val = getattr(self, f, None)
            if val is not None:
                output[f] = val

        return output

    def _self_assign_values(self):
        """
        Assigns values to internal attributes, based on the instance values and falling back to
        values defined at class level. Will also call callables where appropriate.
        """
        for f in self._get_field_list():
            if (
                self.read_only_fields is None or
                (
                    self.read_only_fields is not None and
                    f not in self.read_only_fields
                )
            ):
                instance_val = self._get_value(f)
                if instance_val is not None:
                    setattr(self, f, instance_val)

    def _get_field_list(self):
        """
        If the Meta.fields list is not None we'll use that, otherwise we'll use the list defined
        at class level
        """
        if hasattr(self.Meta, 'fields') and self.Meta.fields is not None:
            fields = self.Meta.fields
        else:
            fields = self.base_fields
        return fields

    def _get_value(self, field_name):
        """
        Gets the value of the field to be assigned to this instance.

        First sees whether the field has a get_X function (where X is the BaseCRM field name) and if
        so, runs it to derive the value from there.

        If that doesn't exist, we check whether the value is specified on this class and if so,
        assign that (calling the method if applicable) -- if not, we see whether the instance has
        an identically-named field and we fall back to the value of that.
        """
        if hasattr(self, 'get_%s' % field_name):
            # we've got a custom method defined to return the value
            val_function = getattr(self, 'get_%s' % field_name)
            val = val_function(self.instance)
        else:
            val = getattr(self.instance, field_name, None)
            if val is not None:
                # it's specified on the instance
                if (
                    isinstance(val, types.FunctionType) or
                    isinstance(val, types.MethodType)
                ):
                    # it's a callable: assign the result
                    val = val()  # note this can easily raise errors - no params sent!
            else:
                # try to get one from the class level
                val = getattr(self, field_name, None)
                if val is not None and isinstance(val, basestring):
                    # it's a string - let's check if it's a field mapping
                    instance_val = getattr(self.instance, val, None)
                    if instance_val is not None:
                        # let's use the mapped value from the instance
                        val = instance_val

        # confirm that whatever we've ended up with is safe to set
        if self._validate_field(field_name, val):
            return val
        else:
            return None

    def _validate_field(self, field_name, field_value):
        """
        Ensures that e.g. owner_id value set is valid for this account; setting an invalid one can
        break the integration as BaseCRM API does not handle this gracefully.

        This abstract method exists to be overridden on concrete serializers
        """
        return True


    class Meta(object):
        model = None
        fields = None


class ContactModelSerializer(AbstractModelSerializer):
    """
    Extends AbstractSerializer. The list of BaseCRM fields for Contact objects exposed via the API
    is visible here: https://developers.getbase.com/docs/rest/reference/contacts
    """
    base_fields = [
        'id',  # BaseCRM ID, read-only
        'owner_id',  # BaseCRM user_id of staffmember assigned to this contact
        'is_organization',
        'contact_id',  # If individual, the ID of the associated organization
        'name',  # Required if is_organization = True
        'first_name',  # Required if is_organization = False
        'last_name',  # Required if is_organization = False
        'customer_status',  # can be 'none', 'current', 'past'
        'prospect_status',  # can be 'none', 'current', 'lost'
        'title',  # job title
        'description',
        'industry',
        'website',
        'email',
        'phone',
        'mobile',
        'fax',
        'twitter',
        'facebook',
        'linkedin',
        'skype',
        'address',  # this is a nested object (dict) with the following keys: 'city', 'country', 'line1', 'postal_code' and 'state'
        'tags',
        'custom_fields',
    ]
    read_only_fields = [
        'id'
    ]

    def _validate_field(self, field_name, field_value):
        if field_value is None:
            return True

        if field_name == 'owner_id' and field_value not in base_helpers.get_user_ids():
            return False

        return super(ContactModelSerializer, self)._validate_field(field_name, field_value)


class DealModelSerializer(AbstractModelSerializer):
    """
    Extends AbstractSerializer. The list of BaseCRM fields for Deal objects exposed via the API
    is visible here: https://developers.getbase.com/docs/rest/reference/deals
    """
    base_fields = [
        'id',  # BaseCRM ID, read-only
        'owner_id',  # BaseCRM user_id of staffmember assigned to this contact
        'name',  # Write-only
        'value',
        'currency',
        'hot',  # Boolean "is it hot?"
        'stage_id',
        'source_id',
        'loss_reason_id',
        'dropbox_email',
        'contact_id',
        'organization_id',  # read-only
        'tags',
        'custom_fields',
    ]
    read_only_fields = [
        'id',
        'organization_id'
    ]

    def _validate_field(self, field_name, field_value):
        if field_value is None:
            return True

        if field_name == 'owner_id' and field_value not in base_helpers.get_user_ids():
            return False

        if field_name == 'stage_id' and field_value not in base_helpers.get_stage_ids():
            return False

        return super(DealModelSerializer, self)._validate_field(field_name, field_value)



# -*- coding: utf-8 -*-
default_app_config = '%s.apps.BaseCRMConfig' % __name__

from helpers import (
    get_contacts,
    get_deals,
    get_notes,
    get_pipelines,
    get_stages,
    create_contact,
    create_deal,
    update_contact,
    update_deal,
)



# -*- coding: utf-8 -*-
import logging
import requests
import urllib

from django.apps import apps as django_apps

from . import (
    settings as base_settings,
    exceptions as base_exceptions
)

logger = logging.getLogger(__name__)

CREATE = '__create__'
UPDATE = '__update__'
RETRIEVE = '__retrieve__'
DELETE = '__delete__'
INFO = '__head__'
VERBS = {
    RETRIEVE: 'GET',
    CREATE: 'POST',
    UPDATE: 'PUT',
    DELETE: 'DELETE',
    INFO: 'HEAD'
}
REQUIRE_BODY_DATA = [
    CREATE,
    UPDATE
]

def request(action, endpoint, get_params=None, **kwargs):
    """
    Makes a request to the BaseCRM API and handles any errors thrown intelligently. See _request
    for more info on the parameters.
    """
    if action not in VERBS.keys():
        raise base_exceptions.BaseCRMBadParameterFormat("Expecting one of INFO, RETRIEVE, CREATE, UPDATE or DELETE but got %s" % action)

    if action in REQUIRE_BODY_DATA:
        if 'data' not in kwargs:
            raise base_exceptions.BaseCRMBadParameterFormat("API method '%s' requires a `data` parameter" % VERBS[action])
        post_data = kwargs.pop('data')
        kwargs['json'] = {'data': post_data}

    if get_params is not None and 'id' in get_params:
        is_id_request = True
    else:
        is_id_request = False

    method = VERBS[action]
    r = _request(method, endpoint, get_params, **kwargs)
    if r.status_code != 200:
        json = r.json()
        logger.error(
            "BaseCRM API responded with status code: '%s'. %s" % (
                r.status_code,
                json['errors'][0]['error']['details'],
            )
        )
        if r.status_code == 401:
            raise base_exceptions.BaseCRMAPIUnauthorized()
        elif r.status_code == 404 and is_id_request:
            # no record with the given ID exists
            raise base_exceptions.BaseCRMNoResult()
        elif r.status_code == 422:
            raise base_exceptions.BaseCRMValidationError(json['errors'][0]['error']['details'])
        else:
            raise Exception(
                "BaseCRM API responded with status code '%s'. %s" % (
                    r.status_code,
                    json['errors'][0]['error']['details'],
                )
            )
    else:
        return r.json()

def parse(response_json):
    """
    Simple helper method to get into the data dicts for each item returned and ignore all the meta;
    just gives back a list of dicts.
    """
    if (
        not isinstance(response_json, dict) or  # wrong type of arg
        (  # ID type response
            'items' not in response_json and
            'data' not in response_json
        ) or
        (  # list endpoint type response
            'items' in response_json and
            (
                len(response_json['items']) > 0 and
                'data' not in response_json['items'][0]
            )
        )
    ):
        raise base_exceptions.BaseCRMBadParameterFormat()

    if 'items' in response_json:
        return [item['data'] for item in response_json['items']]
    else:
        return response_json['data']

def count(response_json):
    """
    Pulls and returns the count given by the BaseCRM API (which takes account of pagination)
    """
    if (
        not isinstance(response_json, dict) or
        'meta' not in response_json
    ):
        raise base_exceptions.BaseCRMBadParameterFormat(
            u'Parameter format was not as expected: a full dict from the requests.json() '
            u'method applied to the full BaseCRM API response. The parameter given was: %s' %
            repr(response_json)
        )

    if 'count' not in response_json['meta']:
        return 1  # we're assuming this is a response to an ID request
    return response_json['meta']['count']

def validate_contact_dict(operation, contact_dict, skip_id=False, suppress=False):
    if suppress == True:
        raise NotImplementedError("No validation suppression in place yet")
    valid = False
    msg = ""
    if operation == CREATE:
        if (
            (  # individual contact requirements
                (
                    'is_organization' not in contact_dict or
                    not contact_dict['is_organization']
                ) and
                (
                    'first_name' in contact_dict and
                    'last_name' in contact_dict
                )
            ) or
            (  # organization contact requirements
                (
                    'is_organization' in contact_dict and
                    contact_dict['is_organization'] == True
                ) and
                (
                    'name' in contact_dict
                )
            )
        ):
            valid = True
        else:
            msg = (
                "If 'is_organization'==True, 'first_name' and 'last_name' are all required, "
                "or if 'is_organization'==False, 'name' is required; "
                "fields supplied were: %s" % contact_dict.keys()
            )
    elif operation == UPDATE:
        if 'id' in contact_dict or skip_id is True:
            valid = True
        else:
            msg = (
                "'id' is required; fields supplied were: %s" % contact_dict.keys()
            )

    if not valid:
        msg = "Parameters fail BaseCRM API requirements for contact. %s" % msg
        raise base_exceptions.BaseCRMValidationError(msg)

    return True

def validate_deal_dict(operation, deal_dict, skip_id=False, suppress=False):
    if suppress == True:
        raise NotImplementedError("No validation suppression in place yet")
    valid = False
    msg = ""
    if operation == CREATE:
        if (
            'name' in deal_dict and
            'contact_id' in deal_dict and
            'custom_fields' in deal_dict
        ):
            valid = True
        else:
            msg = (
                "'name', 'contact_id' and 'custom_fields' are all required; "
                "fields supplied were: %s" % deal_dict.keys()
            )
    elif operation == UPDATE:
        if 'id' in deal_dict or skip_id is True:
            valid = True
        else:
            msg = (
                "'id' is required; fields supplied were: %s" % deal_dict.keys()
            )

    if not valid:
        msg = "Parameters fail BaseCRM API requirements for deal. %s" % msg
        raise base_exceptions.BaseCRMValidationError(msg)

    return True

def instantiate_if_necessary():
    base_app = django_apps.get_app_config('django_basecrm')
    base_app.instantiate_objects()


###########################################################################
##                                                                       ##
##   Don't call any of the following methods directly unless debugging   ##
##                                                                       ##
###########################################################################

def _request(method, endpoint, get_params=None, **kwargs):
    """
    Wraps requests.request with custom headers and api URL generation methods. Accepts just the last
    segment of the endpoint, rather than the whole URI; the rest comes from settings.

    The get_params param will be used to build GET get_params to append to the API call (e.g. for
    pagination and filtering). An ID can also be added in to the get_params param and will be used by
    _build_api_endpoint.

    Any extra kwargs will be passed along to `requests.request()`
    """
    url = _build_api_endpoint(endpoint, get_params)
    extra_headers = kwargs.pop('headers', None)
    kwargs['headers'] = _build_headers(extra_headers)
    kwargs['params'] = get_params

    response = requests.request(method, url, **kwargs)

    logger.debug(
        "'%s' request to BaseCRM API '%s' endpoint gave a '%s' response (url: %s)" % (
            method,
            endpoint,
            response.status_code,
            response.url,
        )
    )
    return response

def _build_headers(extra_headers=None):
    """
    Puts together the standard headers BaseCRM API requires. Accepts a dict that can add to or
    override these.
    """
    headers = {
        'Accept': 'application/json',
        'Content-Type': 'application/json',
        'Authorization': 'Bearer %s' % base_settings.BASECRM_API_KEY,
        'User-Agent': base_settings.BASECRM_USER_AGENT
    }
    if extra_headers is not None:
        if not isinstance(extra_headers, dict):
            raise base_exceptions.BaseCRMBadParameterFormat(
                u'Parameter format was not a dict as expected. The parameter given was: %s' %
                repr(extra_headers)
            )
        for k, v in extra_headers.iteritems():
            headers[k] = v

    return headers

def _build_api_endpoint(endpoint, get_params=None):
    """
    We expect settings.BASECRM_API_URL to be e.g. https://api.getbase.com/v2/ -- note the protocol,
    the path and the trailing slash are all included and expected (i.e. not handled)
    """
    url = '%s%s' % (base_settings.BASECRM_API_URL, endpoint)
    if get_params is not None:
        if not isinstance(get_params, dict):
            raise base_exceptions.BaseCRMBadParameterFormat(
                u'Parameter format was not a dict as expected. The parameter given was: %s' %
                repr(get_params)
            )

        id = get_params.pop('id', None)
        if id is not None:
            url = "%s/%s" % (url, id)
    return url
